@@ -3,9 +3,11 @@ package Net::Whois::Raw;
 require Net::Whois::Raw::Data;
 
 use strict;
-use vars qw($VERSION @ISA @EXPORT @EXPORT_OK $OMIT_MSG $CHECK_FAIL
-	%notfound %strip $CACHE_DIR $CACHE_TIME $USE_CNAMES
-	$TIMEOUT);
+use vars qw(
+    $VERSION @ISA @EXPORT @EXPORT_OK $OMIT_MSG $CHECK_FAIL $CHECK_EXCEED
+    %notfound %strip $CACHE_DIR $CACHE_TIME $USE_CNAMES $TIMEOUT
+    @SRC_IPS
+);
 use IO::Socket;
 
 require Exporter;
@@ -13,26 +15,27 @@ require Exporter;
 @ISA = qw(Exporter);
 
 @EXPORT    = qw( whois whois_config ); ### It's bad manners to export lots.
-@EXPORT_OK = qw( $OMIT_MSG $CHECK_FAIL $CACHE_DIR $CACHE_TIME $USE_CNAMES $TIMEOUT);
+@EXPORT_OK = qw(
+    $OMIT_MSG $CHECK_FAIL $CHECK_EXCEED $CACHE_DIR $CACHE_TIME $USE_CNAMES $TIMEOUT @SRC_IPS );
 
-$VERSION = '0.34';
+$VERSION = '0.43';
 
-($OMIT_MSG, $CHECK_FAIL, $CACHE_DIR, $CACHE_TIME, $USE_CNAMES, $TIMEOUT) = (0) x 6;
+($OMIT_MSG, $CHECK_FAIL, $CHECK_EXCEED, $CACHE_DIR, $CACHE_TIME, $USE_CNAMES, $TIMEOUT) = (0) x 7;
 
 sub whois {
     my ($dom, $srv) = @_;
     my $res;
     unless ($srv) {
-        ($res, $srv) = query($dom);
+	($res, $srv) = query($dom);
     } else {
-        $res = _whois($dom, uc($srv));
+	($res, $srv) = do_whois($dom, uc $srv);
     }
-    finish($res, $srv);
+    return finish($res, $srv);
 }
 
 sub whois_config {
     my ($par) = @_;
-    my @parnames = qw(OMIT_MSG CHECK_FAIL CACHE_DIR CACHE_TIME USE_CNAMES TIMEOUT);
+    my @parnames = qw(OMIT_MSG CHECK_FAIL CACHE_DIR CACHE_TIME USE_CNAMES TIMEOUT @SRC_IPS);
     foreach my $parname (@parnames) {
 	if (exists($par->{$parname})) {
 	    eval('$'.$parname.'='.int($par->{$parname}));
@@ -44,7 +47,7 @@ sub query {
     my $dom = shift;
     my $tld;
     if ($dom =~ /\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}/) {
-        $tld = "IP";
+	$tld = "IP";
     } else { 
 	my @alltlds = keys %Net::Whois::Raw::Data::servers;
 	@alltlds = sort { dlen($b) <=> dlen($a) } @alltlds;
@@ -57,7 +60,7 @@ sub query {
 	}
 	unless ($tld) {
 	    my @tokens = split(/\./, $dom);
-    	    $tld = uc($tokens[-1]); 
+	    $tld = uc $tokens[-1]; 
 	}
     }
 
@@ -70,24 +73,27 @@ sub query {
 	$srv eq 'whois.crsnic.net' ||
 	$srv eq 'whois.publicinterestregistry.net' ||
 	$srv eq 'whois.nic.cc' ||
+	$srv eq 'whois.nic.tv' ||
+	$srv eq 'whois.worldsite.ws' ||
 	$tld eq 'ARPA'
     );
-    my $res = do_whois($dom, uc($srv), $flag, [], $tld);
+    my $res;
+    ($res, $srv) = do_whois($dom, uc $srv, $flag, [], $tld);
     wantarray ? ($res, $srv) : $res;
 }
 
 sub do_whois {
     my ($dom) = @_; # receives 4 parameters, do NOT shift
-    return _whois(@_) unless $CACHE_DIR;
+    return (_whois(@_)) unless $CACHE_DIR;
     mkdir $CACHE_DIR, 0644;
     if (-f "$CACHE_DIR/$dom") {
-        if (open(I, "$CACHE_DIR/$dom")) {
-            my $res = join("", <I>);
-            close(I);
-            return $res;
-        }
+	if (open(I, "$CACHE_DIR/$dom")) {
+	    my $res = join("", <I>);
+	    close(I);
+	    return $res;
+	}
     }
-    my $res = _whois(@_);
+    my ($res) = _whois(@_);
     return $res unless $res;
     return $res unless open(O, ">$CACHE_DIR/$dom");
     print O $res;
@@ -98,31 +104,42 @@ sub do_whois {
 
     my $now = time;
     foreach (glob("$CACHE_DIR/*.*")) {
-        my $atime = (stat($_))[8];
-        my $elapsed = $now - $atime;
-        unlink $_ if ($elapsed / 3600 > $CACHE_TIME); 
+	my $atime = (stat($_))[8];
+	my $elapsed = $now - $atime;
+	unlink $_ if ($elapsed / 3600 > $CACHE_TIME); 
     }
     $res;
 }
 
 sub finish {
     my ($text, $srv) = @_;
-    return $text unless $CHECK_FAIL || $OMIT_MSG;
+    return $text unless $CHECK_FAIL || $OMIT_MSG || $CHECK_EXCEED;
+
+    $srv = lc($srv);
+
+    my $exceed = $Net::Whois::Raw::Data::exceed{$srv};
+    if ($CHECK_EXCEED && $exceed && $text =~ /$exceed/s) {
+	die "Connection rate exceeded";
+    }
+
     *notfound = \%Net::Whois::Raw::Data::notfound;
     *strip = \%Net::Whois::Raw::Data::strip;
 
-    my $notfound = $notfound{lc($srv)};
-    my @strip = $strip{lc($srv)} ? @{$strip{lc($srv)}} : ();
+    my $notfound = $notfound{$srv};
+    my @strip = $strip{$srv} ? @{$strip{$srv}} : ();
     my @lines;
     MAIN: foreach (split(/\n/, $text)) {
-        return undef if $CHECK_FAIL && $notfound && /$notfound/;
-        if ($OMIT_MSG) {
-            foreach my $re (@strip) {
-                next MAIN if (/$re/);
-            }
-        }
-        push(@lines, $_);
+	if ($CHECK_FAIL && $notfound && /$notfound/) {
+	    return undef;
+	};
+	if ($OMIT_MSG) {
+	    foreach my $re (@strip) {
+		next MAIN if (/$re/);
+	    }
+	}
+	push(@lines, $_);
     }
+
     local ($_) = join("\n", @lines, "");
 
     if ($CHECK_FAIL > 1) {
@@ -130,27 +147,27 @@ sub finish {
     }
 
     if ($OMIT_MSG > 1) {	
-        s/The Data.+(policy|connection)\.\n//is;
-        s/% NOTE:.+prohibited\.//is;
-        s/Disclaimer:.+\*\*\*\n?//is;
-        s/NeuLevel,.+A DOMAIN NAME\.//is;
-        s/For information about.+page=spec//is;
-        s/NOTICE: Access to.+this policy.//is;
-        s/The previous information.+completeness\.//s;
-        s/NOTICE AND TERMS OF USE:.*modify these terms at any time\.//s;
-        s/TERMS OF USE:.*?modify these terms at any time\.//s;
-        s/NOTICE:.*expiration for this registration\.//s;
+	s/The Data.+(policy|connection)\.\n//is;
+	s/% NOTE:.+prohibited\.//is;
+	s/Disclaimer:.+\*\*\*\n?//is;
+	s/NeuLevel,.+A DOMAIN NAME\.//is;
+	s/For information about.+page=spec//is;
+	s/NOTICE: Access to.+this policy.//is;
+	s/The previous information.+completeness\.//s;
+	s/NOTICE AND TERMS OF USE:.*modify these terms at any time\.//s;
+	s/TERMS OF USE:.*?modify these terms at any time\.//s;
+	s/NOTICE:.*expiration for this registration\.//s;
 
 	s/By submitting a WHOIS query.+?DOMAIN AVAILABILITY.\n?//s;
 	s/Registration and WHOIS.+?its accuracy.\n?//s;
-        s/Disclaimer:.+?\*\*\*\n?//s;
+	s/Disclaimer:.+?\*\*\*\n?//s;
 	s/The .COOP Registration .+ Information\.//s;
 	s/Whois Server Version \d+\.\d+.//is;
 	s/NeuStar,.+www.whois.us\.//is;
-        s/\n?Domain names in the \.com, .+ detailed information.\n?//s;
-        s/\n?The Registry database .+?Registrars\.\n//s;
-        s/\n?>>> Last update of .+? <<<\n?//;
-        s/% .+?\n//gs;
+	s/\n?Domain names in the \.com, .+ detailed information.\n?//s;
+	s/\n?The Registry database .+?Registrars\.\n//s;
+	s/\n?>>> Last update of .+? <<<\n?//;
+	s/% .+?\n//gs;
     }
     $_;
 }
@@ -161,14 +178,23 @@ sub _whois {
 
     my $sock;
     eval {
-        local $SIG{'ALRM'} = sub { die "Connection timeout to $srv" };
-        alarm $TIMEOUT if $TIMEOUT;
-        $sock = new IO::Socket::INET("$srv:43") || die "$srv: $!";
+	local $SIG{'ALRM'} = sub { die "Connection timeout to $srv" };
+	alarm $TIMEOUT if $TIMEOUT;
+	if (scalar(@SRC_IPS)) {
+	    my $src_ip = $SRC_IPS[0];
+	    push @SRC_IPS, shift @SRC_IPS; # rotate ips
+	    $sock = new IO::Socket::INET(PeerAddr => "$srv:43", LocalAddr => $src_ip) || die "$srv: $!";
+	} else {
+	    $sock = new IO::Socket::INET("$srv:43") || die "$srv: $!";
+	}
     };
     alarm 0;
     die $@ if $@;
     my $israce = $dom =~ /ra--/ || $dom =~ /bq--/;
     my $whoisquery = $dom;
+    if ($srv eq 'WHOIS.CRSNIC.NET') {
+	$whoisquery = "domain $whoisquery";
+    }
     if ($srv eq 'WHOIS.MELBOURNEIT.COM' && $israce) {
 	$whoisquery .= ' race';
     }
@@ -178,38 +204,39 @@ sub _whois {
     close($sock);
     my $answer = join '', @lines;
     if ($flag) {
-        foreach (@lines) {
-            $state ||= (/Registrar:/);
-            if ( $state && /Whois Server:\s*([A-Za-z0-9\-_\.]+)/ ) {
-                my $newsrv = uc("$1");
-                next if (($newsrv) eq uc($srv));
-                return undef if (grep {$_ eq $newsrv} @$ary);
-		my $whois = eval { _whois($dom, $newsrv, $flag, [@$ary, $srv]) };
+	foreach (@lines) {
+	    $state ||= /Registrar/ || /Registered through/;
+	    if ( $state && /Whois Server:\s*([A-Za-z0-9\-_\.]+)/ ) {
+		my $newsrv = uc $1;
+		#warn "recurse to $newsrv\n";
+		next if (($newsrv) eq uc($srv));
+		return undef if (grep {$_ eq $newsrv} @$ary);
+		my ($whois) = eval { _whois($dom, $newsrv, $flag, [@$ary, $srv]) };
 		if ($@ && !$whois || $whois && !check_existance($whois)) {
-		    return $answer;
+		    return ($answer, $srv);
 		}
-                return $whois;
-            }
-            if (/^\s+Maintainer:\s+RIPE\b/ && $tld eq 'ARPA') {
-                my $newsrv = uc($Net::Whois::Raw::Data::servers{'RIPE'});
-                next if ($newsrv eq $srv);
-                return undef if (grep {$_ eq $newsrv} @$ary);
-                my $whois = eval { _whois($dom, $newsrv, $flag, [@$ary, $srv]) };
+		return ($whois, $newsrv);
+	    }
+	    if (/^\s+Maintainer:\s+RIPE\b/ && $tld eq 'ARPA') {
+		my $newsrv = uc $Net::Whois::Raw::Data::servers{'RIPE'};
+		next if ($newsrv eq $srv);
+		return undef if (grep {$_ eq $newsrv} @$ary);
+		my ($whois) = eval { _whois($dom, $newsrv, $flag, [@$ary, $srv]) };
 		if ($@ && !$whois) {
-		    return $answer;
+		    return ($answer, $srv);
 		}
-                return $whois;
-            }
-        }
+		return ($whois, $newsrv);
+	    }
+	}
     }
     my $whois = join("", @lines);
 
     if ($whois =~ /To single out one record, look it up with \"xxx\",/s) {
-        my $newsrv = uc('whois.networksolutions.com');
-        return _whois($dom, $newsrv, $flag, [@{$ary||[]}, $srv]);
+	#my $newsrv = uc 'whois.networksolutions.com';
+	return _whois( "=$dom", $srv, $flag, [@{$ary||[]}, $srv] );
     }
     
-    return $whois;
+    return ($whois, $srv);
 }
 
 sub dlen {
@@ -236,8 +263,9 @@ sub check_existance {
 	/No domain records were found/s ||
 	/No such domain/s ||
 	/No entries found in the /s ||
+	/Could not find a match for/s ||
 	/Unable to find any information for your query/s ||
-	/is not registered and may be available for registration/s;
+	/is not registered/s;
     return 1;
 }
 
@@ -262,9 +290,10 @@ Net::Whois::Raw - Perl extension for unparsed raw whois information
   $s = whois('funet.fi');
   $s = whois('yahoo.co.uk');
 
-	### if you do "use Net::Whois::Raw qw( whois $OMIT_MSG $CHECK_FAIL 
-	###              $CACHE_DIR $CACHE_TIME $USE_CNAMES $TIMEOUT );
-	### you can use these:
+  ### if you do "use Net::Whois::Raw qw(
+  #		$OMIT_MSG $CHECK_FAIL $CHECK_EXCEED
+  #	    	$CACHE_DIR $CACHE_TIME $USE_CNAMES $TIMEOUT @SRC_IPS );
+  ### you can use these:
 
   $OMIT_MSG = 1; # This will attempt to strip several known copyright
 		messages and disclaimers sorted by servers.
@@ -281,6 +310,10 @@ Net::Whois::Raw - Perl extension for unparsed raw whois information
   $CHECK_FAIL = 2; # This will match against several more rules
 		if none are known for the specific server.
 
+  $CHECK_EXCEED = 1; # When this option is set, "die" will be called
+		is connection rate to specific whois server have been
+		exceeded
+
   $CACHE_DIR = "/var/spool/pwhois/"; # Whois information will be
 		cached in this directory. Default is no cache.
 
@@ -297,8 +330,7 @@ Net::Whois::Raw - Perl extension for unparsed raw whois information
   $TIMEOUT = 10; # Cancel the request if connection is not made within
 		a specific number of seconds.
 
-  Note: as of version 0.21, extra data will be loaded only if the
-  OMIT_MSG or CHECK_FAIL flags were used, in order to reduce memory usage.
+  @SRC_IPS = (11.22.33.44); # List of local IP addresses to
 
 =head1 DESCRIPTION
 
